/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   BSD - specific code for x86 processors.
 *   this file is a part of xtrace project
 *
 * $Id: bsd-i386.c,v 1.12 2002/09/23 14:57:17 drozd Exp $
 */

#include <machine/reg.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#include <stdio.h>
#include <signal.h>

#include "xtrace.h"
#include "bsd.h"

/*
 * Get call data by stack pointer
 */
int
xtr_get_call_data(struct opt *op, struct calldata *cdp)
{
	static vaddr_t args[MAXARGS], *argsp[MAXARGS];
	struct bsddata *dp = (struct bsddata *) op->data;
	static struct reg regs;
	vaddr_t esp;
	int i, scno;

	for (i = 0; i < sizeof(argsp)/sizeof(argsp[0]); i++)
		argsp[i] = args + i;

	if (dp->regsfd < 0)
		return (-1);

	if (lseek(dp->regsfd, 0L, SEEK_SET) == -1)
		perror("xtr_get_call_data: lseek()");

	if (read(dp->regsfd, &regs, sizeof(regs)) != sizeof(regs)) {
		perror("xtr_get_call_data: read(/proc/pid/regs)");
		return (-1);
	}

	esp = regs.r_esp + sizeof(int);
		
	cdp->retval = regs.r_eax;
	cdp->argv = argsp;
	cdp->ip = regs.r_eip;

	if (op->flags & XTR_INSYSENTER) {
		
		op->scno = regs.r_eax;
		
		/* handle indirect syscalls */
		if (op->scno == SYS___syscall || op->scno == SYS_syscall) {

			(void) lseek(dp->procfd, esp, SEEK_SET);
			(void) read(dp->procfd, &scno, sizeof(int));

			if (op->scno == SYS_syscall)
				/* int - aligned */
				esp += sizeof(int);
			else
				/* quad - aligned */
				esp += sizeof(quad_t);

			op->scno = scno;

			dprintf((stderr, "*** xtr_get_call_data:"
				       " indirect syscall %d\n", op->scno));
		}
	}


	/*
	 * in the case of backtrace or function call, limit the
	 * number of arguments
	 */
	if (!(op->flags & XTR_INSYSENTER) || (op->flags & XTR_BACKTRACING))
		argsp[4] = NULL;

	/*
	 * When stopped at function prologue, the frame pointer is not set yet
	 */
	if (cdp->fp == 0U) {
		if (BIT_ISSET(op->flags, XTR_INFENTER)) {
			esp =  cdp->fp = regs.r_esp;
		} else {
			cdp->fp = regs.r_ebp;
		}
	}

	if (!(op->flags & XTR_INSYSEXIT) || op->flags & XTR_BACKTRACING) {
		
		/* read MAXARGS arguments from stack */
		if (op->flags & XTR_INSYSENTER)
			(void) lseek(dp->procfd, esp, SEEK_SET);
		else if (op->flags & XTR_INFENTER)
			(void) lseek(dp->procfd, esp + 4, SEEK_SET);
		else
			(void) lseek(dp->procfd, cdp->fp + 8, SEEK_SET);

		(void) read(dp->procfd, args, sizeof(args));

		if (cdp->fp != 0) {
			(void) xtr_read_mem(op, cdp->fp, &cdp->ofp,
			    sizeof(vaddr_t));
			(void) xtr_read_mem(op, cdp->fp +4, &cdp->rta,
			    sizeof(vaddr_t));
		}
	}

	return (0);
}

/*
 * Handle the signal child received
 */
void
handle_signal(struct opt *op)
{
	struct bsddata *dp = (struct bsddata *) op->data;
	struct symbol *symp;
	struct reg regs;
#ifdef __DEBUG__
	unsigned int instr;
#endif /* __DEBUG__ */

	dprintf((stderr, "*** handle_signal (sig %d)\n", op->signo));

	if (BIT_ISSET(op->flags, XTR_STEPPING)) {
		BIT_UNSET(op->flags, XTR_STEPPING);
		dprintf((stderr, "*** handle_signal: stepping\n"));
		goto out;
	}

	if (op->signo != SIGTRAP) {
		op->flags |= XTR_SIGNAL;
		goto dump;
	}

	op->flags |= XTR_INFENTER;

	/*
	 * fetch the registers to find out where the child is stopped
	 */
	if (lseek(dp->regsfd, 0, SEEK_SET) == -1) {
		dprintf((stderr, "*** handle_signal: lseek(regs) failed"));
		goto out;
	}
	if (read(dp->regsfd, &regs, sizeof(regs)) != sizeof(regs)) {
		dprintf((stderr, "*** handle_signal: read(regs) failed"));
		goto out;
	}
	
	/* EIP is known now, so find the appropriate symbol  */
	if ((symp = xtr_find_symbol(regs.r_eip, 0)) == NULL) {
		dprintf((stderr, "*** no symbol for address %#x\n", regs.r_eip));
		goto out;
	}
#ifdef __DEBUG__
	(void) xtr_read_mem(op, regs.r_eip - 1, &instr, sizeof(instr));
	dprintf((stderr, "*** bpt hit %s (%#x, EIP %#x -> %#x)\n",
		       symp->name, symp->va, regs.r_eip, instr));
#endif /* __DEBUG__ */

	/*
	 * restore old instruction
	 */
	if (xtr_clear_bpt(op, regs.r_eip - 1) != 0)
		goto out;

	/*
	 * SIGTRAP has been sent _after_ INT3 instruction, so we have to
	 * move the instruction pointer back 1 byte after we've restored
	 * the old instruction
	 */
	--regs.r_eip;
	if (write(dp->regsfd, &regs, sizeof(regs)) != sizeof(regs))
		perror("xtr_get_call_data: write(regs)");

	/*
	 * set the single-step flag and perform single step.
	 * this control command will just set the TF (trap flag) in i386
	 * EFLAGS register, so SIGTRAP will occur after the execution of
	 * restored instruction. we will ignore SIGTRAP in the signal handler
	 * if single step flag set.
	 */
	op->flags |= XTR_STEPPING;
	ctl(dp->ctlfd, "step");

	/*
	 * at this point signal handler must receive SIGTRAP, ignore it and
	 * clear single-step flag. install breakpoint again, and continue
	 * the child. The program itself stays two instructions forward after
	 * beginning of the fuction, but 'regs' structure contains the old
	 * value of EIP register, that points at the function entrance. So
	 * use it to set a breakpoint.
	 */
	xtr_set_bpt(op, regs.r_eip);

dump:

	xtr_dump_call(op);

out:
	/* continue the debuggee */
	ctl(dp->ctlfd, "run");
}
