/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   BSD - specific code
 *   this file is a part of xtrace project
 *
 * $Id: bsd.c,v 1.12 2002/09/25 10:49:42 drozd Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

#include "xtrace.h"
#include "bsd.h"
#include "elf.h"

static struct bsddata __data;		/* bsd - specific data */
static void fini(struct opt *);

int
xtr_stop_app(struct opt *op)
{
	struct bsddata *dp = (struct bsddata *) op->data;
	const char *reason;

	dprintf((stderr, "*** xtr_stop_app: stopping app... "));

	if (ioctl(dp->procfd, PIOCWAIT, &dp->status) < 0) {
		perror("ioctl(PIOCWAIT)");
		return (-1);
	}

	switch (dp->status.why) {
		
		case S_SCE: 
			reason = "sysenter";
			SET_REASON(op, XTR_INSYSENTER);
			break;
		
		case S_EXEC:
			reason = "execve";
			BIT_SET(op->flags, XTR_INEXEC);
			break;
		
		case S_SCX:
			reason = "sysexit";
			SET_REASON(op, XTR_INSYSEXIT);
			break;
		
		case S_SIG:
			reason = "signal";
			SET_REASON(op, XTR_SKIPPING);
			op->signo = dp->status.val;
			break;
		
		default:
			    reason = "unknown";
			    SET_REASON(op, XTR_OTHER);
			    break;
	}

	dprintf((stderr, "[%s]\n", reason));
				
	if (dp->status.why == S_EXIT) {
		BIT_SET(op->flags, XTR_CHILDEXITED);
		exit(1);
	}

	return (0);
}

int
xtr_resume_app(struct opt *op)
{
	struct bsddata *dp = (struct bsddata *) op->data;
	int val = 0;

	dprintf((stderr, "*** xtr_resume_app: resuming\n"));

	if (ioctl(dp->procfd, PIOCCONT, val) < 0) {
		perror("ioctl(PIOCCONT)");
		return (-1);
	}

	return (0);
}

/*
 * Initialize tracer
 * @op - current options
 * @slist - comma separated list of syscalls to trace
 * @flist - comma separated list of functions to trace
 * Return:  0 on success, < 0 on error
 */
int
xtr_init(struct opt *op)
{
	char fname[128];
	struct bsddata *bdp = &__data;
	struct procfs_status status;
	unsigned long flags;

	/* register cleanup and SIGCLD handler functions, setup data ptr */
	bdp->op = op;
	op->fini = fini;
	op->data = &__data;
	op->child = handle_signal;

	/*
	 * open a ctl, regs, mem  files. Remember old event mask.
	 * set new event mask
	 */

	(void) snprintf(fname, sizeof(fname), "/proc/%d/mem", op->pid);
	if ((bdp->procfd = open(fname, O_RDWR)) < 0) {
		perror("xtr_init: open(/proc/pid/mem, O_RDWR"); exit(1);
	}

	flags = fcntl(bdp->procfd, F_GETFL);
	flags |= FD_CLOEXEC;
	(void) fcntl(bdp->procfd, F_SETFL, flags);

	(void) snprintf(fname, sizeof(fname), "/proc/%d/regs", op->pid);
	if ((bdp->regsfd = open(fname, O_RDWR)) < 0) {
		perror("xtr_init: open(/proc/pid/regs, O_RDWR"); exit(1);
	}
	
	(void) snprintf(fname, sizeof(fname), "/proc/%d/ctl", op->pid);
	if ((bdp->ctlfd = open(fname, O_WRONLY)) < 0) {
		perror("xtr_init: open(ctl)"); exit(1);
	}

	if (ioctl(bdp->procfd, PIOCSTATUS, &status) == -1) {
		perror("xtr_init: ioctl(PIOCSTATUS)"); exit(1);
	}
	bdp->oevnts = status.events;

	/* clear PF_LINGER flag */
	if (ioctl(bdp->procfd, PIOCGFL, &flags) == -1) {
		perror("xtr_init: ioctl(PIOCGFL)"); exit(1);
	}
	flags &= ~PF_LINGER;
	if (ioctl(bdp->procfd, PIOCSFL, flags) == -1) {
		perror("xtr_init: ioctl(PIOCSFL)"); exit(1);
	}
	
	bdp->evnts = S_SCE | S_SCX | S_EXIT | S_SIG | S_EXEC;
	if (ioctl(bdp->procfd, PIOCBIS, bdp->evnts) == -1) {
		perror("xtr_init: ioctl(PIOCBIS)"); exit(1);
	}

	(void) ctl(bdp->ctlfd, "attach");
	
	return (0);
}

/*
 * Dump memory content from process's address space
 */
int
xtr_read_mem(const struct opt *op, vaddr_t va, void *buf, int len)
{
	int ret;
	struct bsddata *dp = (struct bsddata *) op->data;

	if (lseek(dp->procfd, va, SEEK_SET) == -1) {
		dprintf((stderr,
		    "*** xtr_read_mem: lseek(va): errno %d\n", errno));
		exit(1);
	}
	
	if ((ret = read(dp->procfd, buf, len)) != len)
		dprintf((stderr,
		    "*** xtr_read_mem: read(va): errno %d\n", errno));

	return (ret);
}

/*
 * Write into process's address space
 */
int
xtr_write_mem(const struct opt *op, vaddr_t va, const void *buf, int len)
{
	int ret;
	struct bsddata *dp = (struct bsddata *) op->data;

	if (lseek(dp->procfd, va, SEEK_SET) == -1) {
		dprintf((stderr,
		    "*** xtr_write_mem: lseek(va): %s\n",strerror(errno)));
		exit(1);
	}
	
	if ((ret = write(dp->procfd, buf, len)) != len)
		dprintf((stderr,
		    "*** xtr_write_mem: read(va): %s\n",strerror(errno)));

	return (ret);
}

/*
 * Cleanup function. Must handle Linux-specific stuff.
 * See what xtr_fini() does before patching this code
 */
void
fini(struct opt *op)
{
	struct bsddata *bdp = (struct bsddata *) op->data;
		
	if (!(op->flags & XTR_CHILDEXITED))
		if (ioctl(bdp->procfd, PIOCBIS, bdp->oevnts) == -1) 
			dprintf((stderr, "*** fini: ioctl(PIOCBIS)\n"));

	(void) ctl(bdp->ctlfd, "detach");

	(void) close(bdp->procfd);
	(void) close(bdp->regsfd);
	(void) close(bdp->ctlfd);
}

/*
 * Send specified command `cmd' to a /proc/pid/ctl procfs control file
 */
int
ctl(int fd, const char *cmd)
{
	int ret = 0;
	int len = strlen(cmd);

	if (write(fd, cmd, len) != len) {
		dprintf((stderr, "*** ctl: %s (%s)\n", cmd, strerror(errno)));
		--ret;
	}

	return (ret);
}
