/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   ELF manipulation routines needed by xtrace utility, see README for details.
 *   this file is a part of xtrace project
 *
 * $Id: elf.c,v 1.10 2002/09/25 09:17:23 drozd Exp $
 */

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/param.h>

#ifdef FREEBSD
#include <machine/vmparam.h>
#endif /* FREEBSD */

#ifdef SVR4
#include <sys/link.h>
#endif /* SVR4 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>

#include "elf.h"

/* load address of memory-mapped ELF binary */
static void *__loadaddr;

/* struct stat for that memory-mapped ELF binary */
static struct stat __st;

static Elf32_Dyn *find_dynamic_section(const Elf32_Ehdr *, Elf32_Sword);
static const char *abspath(const char *fname);

/*
 * Open an ELF image, memory map it and return ptr to an ELF header
 * if `action' parameter is 0, then unmap existing mapping
 * Parameters:
 *   fd - file descriptor of an ELF object to open
 *   action - if 1, map it, of 0, then unmap file.
 * Return:
 *   ptr to Elf header. it equals to mapping's load address
 */
Elf32_Ehdr *
elf_init(int fd)
{
	/* are we free ? */
	if (__loadaddr != 0) {
		(void) fprintf(stderr, "__loadaddr != 0, call elf_fini() first\n");
		elf_fini(__loadaddr);
	}

	/* find the size */
	if (fstat(fd, &__st) != 0) {
		perror("elf_open fstat()");
		return 0;
	}

	__loadaddr = mmap(NULL, __st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);

	if (__loadaddr == MAP_FAILED) {
		perror("elf_open mmap()");
		__loadaddr = 0;
	}

	return (Elf32_Ehdr *) __loadaddr;
}

/*
 * Close ELF mapping
 */
void
elf_fini(Elf32_Ehdr *ep)
{
	if (ep != __loadaddr) {
		(void) fprintf(stderr, "race, ep != __loadaddr\n");
	}

	(void) munmap((void *) __loadaddr, __st.st_size);
	__loadaddr = 0;
}

/*
 * Return the ptr to character string for given offset in string table
 * Parameters:
 *   ep - ptr to ELF object
 *   secindex - section header index with a string info
 *   offset - offset into string table for given object
 * Return:
 *   ptr to a string
 */
const char *
elf_str(const Elf32_Ehdr *ep, int secindex, int offset)
{
	char *stable;
	Elf32_Shdr *shdrp;

	if (ep == NULL || secindex > ep->e_shnum)
		return 0;

	shdrp = ((Elf32_Shdr *) ((char *) ep + ep->e_shoff)) + secindex;
	stable = (char *) ep + shdrp->sh_offset;

	return (stable + offset);
}

/*
 * Return the ptr to section header for given index
 */
Elf32_Shdr *
elf_getshdr(const Elf32_Ehdr *ep, int index)
{
	if (ep == NULL || index > ep->e_shnum)
		return 0;

	return (((Elf32_Shdr *) ((char *) ep + ep->e_shoff)) + index);
}

/*
 * Find the section index with the given name
 * Parameters:
 *   ep - Elf load address
 *   name - section's name
 *   out - section ptr will be put there
 * Return:
 *   index of a section, or -1 on failure
 */
int
elf_getindex(const Elf32_Ehdr *ep, const char *name, Elf32_Shdr **out)
{
	Elf32_Shdr *shdrp;
	int i, found;
	const char *secname;

	/* find the section table start */
	shdrp = elf_getshdr(ep, 0);

	for (i = 0, found = 0; i < ep->e_shnum; i++) {
		secname = elf_str(ep, ep->e_shstrndx, shdrp[i].sh_name);
		if (!strcmp(name, secname)) {
			++found;
			break;
		}
	}

	/* pass the pointer upstairs */
	if (out != NULL)
		*out = shdrp + i;

	return (found > 0 ? i : -1);
}

/*
 * Return load address for specified ELF object
 * Parameters:
 *    op - ptr to the options
 *    fname - ELF object's file name
 * Return:
 *    load address
 */
vaddr_t
elf_get_load_addr(const Elf32_Ehdr *ep)
{
	Elf32_Phdr *phdrp;
	vaddr_t retval;
	int k;

	if (ep == 0)
		return 0;

	/*
	 * SVR4 ABI says:
	 * ``To compute the base address, one determines the memory address
	 * associated with the lowest p_vaddr value for a PT_LOAD segment''
	 * we will use another algorythm - find PT_LOAD section with
	 * write permission turned on, and get p_vaddr
	 */

	/* find the program header table */
	phdrp = (Elf32_Phdr *) ((char *) ep + ep->e_phoff);

	/* iterate over all program headers */
	for (k = 0, retval = 0; k < ep->e_phnum; k++) {
		if (phdrp[k].p_type == PT_LOAD){
			retval = retval == 0
				? phdrp[k].p_vaddr
				: MIN(retval, phdrp[k].p_vaddr); 
		}
	}

	if (retval == (vaddr_t) -1) {
		(void) fprintf(stderr, "invalid ELF: wrong PT_LOAD segments\n");
		return 0;
	}

	return retval;
}

/*
 * Get list of dependencies for the given ELF file
 * Parameters:
 *   ep - ptr to ELF header
 * Return:
 *   list of file names
 */
char *const *
elf_get_depends(const Elf32_Ehdr *ep)
{
	static char *fnames[128];
	Elf32_Dyn *dynp;
	const char *strtab;
	int i;
	const char *fname;

	if (ep == 0)
		return 0;

	/* find load address */
	if ((dynp = find_dynamic_section(ep, DT_STRTAB)) == 0)
		return 0;

	strtab = (char *) ep + (dynp->d_un.d_ptr - elf_get_load_addr(ep));

	if (strtab == NULL) {
		(void) fprintf(stderr, "no DT_STRTAB entry\n");
		return 0;
	}

	dynp = find_dynamic_section(ep, DT_NEEDED);

	/* iterate over dynamic sections */
	for (i = 0; dynp != NULL && dynp->d_tag != DT_NULL; dynp++) {
		if (dynp->d_tag == DT_NEEDED) {
			fname = abspath(strtab + dynp->d_un.d_val);
			if (fname != 0) {
				fnames[i++] = strdup(fname);
			} else {
				(void) fprintf(stderr, "cannot find %s\n",
				    (strtab + dynp->d_un.d_val));
			}
		}
	}

	/*
	 * TODO: add interpreter as well
	 * Solaris has dynamic linker in dependency list
	 */

	/* NULL - terminate the list */
	fnames[i++] = 0;

	return fnames;
}

/*
 * Find dynamic section by the given tag
 */
Elf32_Dyn *
find_dynamic_section(const Elf32_Ehdr *ep, Elf32_Sword tag)
{
	Elf32_Dyn *dynp, *retval;
	Elf32_Phdr *phdrp;
	int i;

	if (ep == 0)
		return 0;

	/* find PT_DYNAMIC section */
	phdrp = (Elf32_Phdr *) ((char *) ep + (int) ep->e_phoff);
	for (i = 0; i < ep->e_phnum; i++)
		if (phdrp[i].p_type == PT_DYNAMIC)
			break;

	if (phdrp[i].p_type != PT_DYNAMIC) {
		dprintf((stderr, "*** find_dynamic_section: wrong ELF\n"));
		return 0;
	}

	dynp = (Elf32_Dyn *) ((char *) ep + (int) phdrp[i].p_offset);

	/* skip first entry - it is always DT_NULL */
	/*
	++dynp;
	*/

	/* find string table */
	for (retval = 0; dynp->d_tag != DT_NULL; dynp++) {
		if (dynp->d_tag == tag) {
			retval = dynp;
			break;
		}
	}
	
	return retval;
}


/*
 * Find absolute file name for given ELF file name
 * */
const char *
abspath(const char *fname)
{
	const char *const paths[] = { "/usr/lib", "/usr/local/lib", 
	    "/lib", "/usr/X11R6/lib", "/usr/openwin/lib", NULL };
	static char try[128], link[512];
	const char *const *path;
	const char *retval = 0;
	struct stat st;
	int nbytes;
	char *p;

	/* try to find absolute filename */
	if (access(fname, R_OK) == 0) {
		retval = fname;
	} else {
		for (path = paths; *path != NULL; path++) {
			(void) snprintf(try, sizeof(try), "%s/%s",
			    *path, fname);

			if (access(try, R_OK) == 0) {
				retval = try;
				break;
			}
		}
	}

	/* check is it a soft link */
	if ((retval != 0) && lstat(retval, &st) == 0) {
		if (S_ISLNK(st.st_mode)) {
			(void) strcpy(link, retval);
			p = strrchr(link, '/');
			if (fname != 0) {
				nbytes = p - link;
				++nbytes;
				++p;
				nbytes += readlink(retval, p, sizeof(link));
			} else {
				nbytes = readlink(retval, link, sizeof(link));
			}
			link[nbytes] = '\0';
			retval = link;
		}
	}

	return retval;
}

/*
 * Returs 1 if executable is statically linked, 0 otherwise
 */
int
elf_isstatic(const Elf32_Ehdr *ep)
{
	int ret = 0;

	if (find_dynamic_section(ep, DT_NEEDED) == NULL)
		++ret;

	return (ret);
}

/*
 * return path to dynamic linker. path is statically allocated
 */
void
elf_dynloader(const Elf32_Ehdr *ep, char **path, vaddr_t *base)
{
	Elf32_Phdr *phdrp;
	int i;

	if (path != NULL)
		*path = NULL;
	
	phdrp = (Elf32_Phdr *) ((char *) ep + (int) ep->e_phoff);
	
	/* find PT_INTERP section */
	for (i = 0; i < ep->e_phnum; i++)
		switch (phdrp[i].p_type) {
			
			case PT_INTERP:
				if (path != NULL)
					*path = strdup((char *) ep +
					    phdrp[i].p_offset);
				break;
			
			case PT_LOAD:
				/*
				 * The dynamic linker is mapped to an address
				 * that equals
				 * p_vaddr + maximal data segment size.
				 */
#ifdef FREEBSD
				if (base != NULL && (phdrp[i].p_flags & PF_W))
					*base = trunc_page(phdrp[i].p_vaddr)
					    + MAXDSIZ;
#endif /* FREEBSD */
				break;
			default:
				break;
		}

}
