/*
 * Copyright (c) 2002 Silver Software Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   Linux, i386 - specific code
 *   this file is a part of xtrace project
 *
 * $Id: linux-i386.c,v 1.6 2002/09/25 08:33:48 drozd Exp $
 */


#include <sys/ptrace.h>
#include <sys/syscall.h>
#include <linux/user.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "xtrace.h"
#include "linux.h"

static struct user_regs_struct *getregs(struct opt *);
static int is_trap(struct opt *);
static void handle_trap(struct opt *);

/*
 * Get general purpose registers
 */
struct user_regs_struct *
getregs(struct opt *op)
{
	static struct user_regs_struct regs, *ret = &regs;
	
	/* copy the registers */
	while (ptrace(PTRACE_GETREGS, op->pid, 0, &regs) != 0) {
		if (!xtr_isalive(op->pid)) {
			BIT_SET(op->flags, XTR_CHILDEXITED);
			ret = NULL;
			break;
		}
	}

	return (ret);

}

/*
 * Determine - is the current stop because of SIGTRAP or because of syscall
 * or single step. do this by examining a byte before program counter - if
 * it 0xCC (INT3), then child is stopped because of SIGTRAP, otherwise not.
 * Return: 1 if it is a trap, otherwise 0
 */
int
is_trap(struct opt *op)
{
	struct user_regs_struct *rp;
	unsigned char ch;
	int ret = 0;

	if ((rp = getregs(op)) != NULL) {
		
		/* read the instruction just before EIP */
		(void) xtr_read_mem(op, rp->eip - 1, &ch, sizeof(ch));
		
		/* if it is BPT instruction, return 1 */
		if (ch == BPT[0])
			++ret;
	}
	
	dprintf((stderr, "*** is_trap: %s, eip: %lx, code: %x\n",
	    ret ? "yes" : "no", rp->eip, ch));

	return (ret);
}

/*
 * Get call data by stack pointer
 * this function must set the following fields:
 *   o  op->scno
 *   o  all fields from calldata structure
 */
int
xtr_get_call_data(struct opt *op, struct calldata *cdp)
{
	struct user_regs_struct *rp;
	static vaddr_t *argv[MAXARGS], args[MAXARGS];
	int i;

	if (cdp == 0)
		return (-1);

	args[0] = 0;

	if (argv[0] == NULL)
		for (i = 0; i < sizeof(argv)/sizeof(argv[0]); i++)
			argv[i] = &args[i];

	if ((rp = getregs(op)) == 0)
		return (-1);

	/* fill up call data */
	cdp->ip = rp->eip;

	if (cdp->fp == 0U) 
		cdp->fp = BIT_ISSET(op->flags, XTR_INFENTER) ? rp->esp: rp->ebp;

	(void) xtr_read_mem(op, cdp->fp, &cdp->ofp, sizeof(cdp->ofp));
	(void) xtr_read_mem(op, cdp->fp + 4, &cdp->rta, sizeof(cdp->rta));
	
	cdp->argv = argv;
	cdp->retval = rp->eax;

	op->scno = rp->orig_eax;

	if ((op->flags & XTR_INSYSENTER) || (op->flags & XTR_INSYSEXIT)) {
		/* parameters are in registers */
		args[0] = rp->ebx; args[1] = rp->ecx; args[2] = rp->edx;
		args[3] = rp->esi; args[4] = rp->edi; args[5] = rp->ebp;
		args[6] = 0;
	} else if (BIT_ISSET(op->flags, XTR_INFENTER)) {
		(void) xtr_read_mem(op, rp->esp + sizeof(vaddr_t),
		    args, sizeof(args));
	}

	
	return (0);
}


void
handle_signal(struct opt *op)
{
	struct user_regs_struct *rp;

	dprintf((stderr, "*** handle_signal:\n"));
	dumpregs(op);

	if (BIT_ISSET(op->flags, XTR_STEPPING)) {
		
		BIT_UNSET(op->flags, XTR_STEPPING);
		SET_REASON(op, XTR_SKIPPING);
		dprintf((stderr, "*** handle_signal: single step. bp reset on"
		    " va %#x\n", op->stopva));
		(void) xtr_set_bpt(op, op->stopva);
		return;
	}

	switch (op->signo) {
	
		case SIGTRAP:
			
			if (is_trap(op)) {
				/* is it a function entrance ? */
				BIT_SET(op->flags, XTR_STEPPING);
				SET_REASON(op, XTR_INFENTER);
				handle_trap(op);

			} else {
				/* is it a syscall enter/exit ? */
				rp = getregs(op);
				if (rp == NULL)
					exit(1);

				if (rp->eax == -ENOSYS) {
					SET_REASON(op, XTR_INSYSENTER);
					
					/* if child exits  */
					if (rp->orig_eax == SYS_exit)
						BIT_SET(op->flags,
						    XTR_CHILDEXITED);
				} else {
					SET_REASON(op, XTR_INSYSEXIT);
					
					/* set inexec flag, if in exec */
					if (rp->orig_eax == SYS_execve)
						BIT_SET(op->flags, XTR_INEXEC);
					
				}
			}
			
			break;
		
		default:
			SET_REASON(op, XTR_SIGNAL);
			break;
	}
}


/*
 * Handle the breakpoint signal. We have to restore the old instruction, then
 * single step over it, then set the breakpoint again and continue.
 */
void
handle_trap(struct opt *op)
{
	struct user_regs_struct *rp;

	dprintf((stderr, "*** handle_trap\n"));

	if ((rp = getregs(op)) == NULL)
		return;
	
	/* move one instruction back */
	op->stopva = --rp->eip;

	/* clear the breakpoint */
	(void) xtr_clear_bpt(op, rp->eip);

	/* save registers */
	if (ptrace(PTRACE_SETREGS, op->pid, 0, rp) != 0)
		perror("*** setregs: PTRACE_SETREGS");
}

/*
 * dump registers
 */
void
dumpregs(struct opt *op)
{
	struct user_regs_struct *rp;

	if ((rp = getregs(op)) != NULL) {
		dprintf((stderr, "*** dumpregs:"
		    " eip: %lx eax: %lx orig_eax: %lx\n",
		    rp->eip, rp->eax, rp->orig_eax));
	}
}
