/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   Linux - specific code
 *   this file is a part of xtrace project
 *
 * $Id: linux.c,v 1.9 2002/09/25 10:49:42 drozd Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

#include "xtrace.h"
#include "linux.h"
#include "elf.h"

static struct linuxdata __data;		/* linux - specific data */

static void fini(struct opt *);


/*
 * Initialize tracer
 * Parameters:
 *    op - current options
 * Return:
 *    0 on success, < 0 on error
 */
int
xtr_init(struct opt *op)
{
	char fname[64];

	/* register cleanup and SIGCLD handler functions, setup data ptr */
	op->fini = fini;
	op->data = &__data;
	op->child = handle_signal;
	__data.op = op;

	/* open memory */
	(void) snprintf(fname, sizeof(fname), "/proc/%d/mem", op->pid);
	if ((__data.memfd = open(fname, O_RDONLY)) == -1) {
		perror("xtr_init: open(/proc/pid/mem)");
		return (-1);
	}

	/* attach to a process */
	if (ptrace(PTRACE_ATTACH, op->pid, (char *) 1, 0) != 0) {
		perror("ptrace(PTRACE_ATTACH)");
		return (-1);
	}

	return (0);
}

/*
 * Stop on event of interest
 */
int
xtr_stop_app(struct opt *op)
{
	
	dprintf((stderr, "*** xtr_stop_app: locking, flags %#x\n", op->flags));

	/* wait for the lock */
	LOCK(op);
	
	dprintf((stderr, "*** xtr_stop_app: got lock, flags %#x\n", op->flags));


	dumpregs(op);

	return (0);
}

/*
 * Resume process
 */
int
xtr_resume_app(struct opt *op)
{
	int state, ret = 0;
	
	dprintf((stderr, "*** xtr_resume_app: continuing child, flags %#x)\n",
	    op->flags));

	if (BIT_ISSET(op->flags, XTR_STEPPING)) {
		if (ptrace(PTRACE_SINGLESTEP, op->pid, 0, 0)) {
			dprintf((stderr, "*** handle_trap:
			    ptrace(PTRACE_SINGLESTEP) [%s]", strerror(errno)));
			ret--;
		}
		state = XTR_STEPPING;
	} else if (ptrace(PTRACE_SYSCALL, op->pid, 0, 0) == -1) {
		dprintf((stderr, "*** xtr_resume_app: "
		    "ptrace(PTRACE_SYSCALL) (%s)", strerror(errno)));
		--ret;
	}
	
	dprintf((stderr, "*** xtr_resume_app: ptrace sent (%s)\n",
	    state == XTR_STEPPING ? "singlestep" : "normal"));


	return (ret);
}

/*
 * Cleanup function. Must handle Linux-specific stuff.
 * See what xtr_fini() does before patching this code
 */
void
fini(struct opt *op)
{
	dprintf((stderr, "*** fini: detaching from a process\n"));

	if (ptrace(PTRACE_DETACH, op->pid, 0, 0) != 0)
		perror("ptrace(PTRACE_DETACH)");
	
	(void) putchar('\n');
}

/*
 * Read memory block from debuggee's address space
 */
int
xtr_read_mem(const struct opt *op, vaddr_t from, void *to, int len)
{
	int nbytes, n;
	size_t i;
	unsigned long word;
	unsigned char *ucp, *dest;

	dest = (unsigned char *) to;

	for (n = len/sizeof(word) + 1, nbytes = 0;
	    n != 0; n--, from += sizeof(long)) {
		word = ptrace(PTRACE_PEEKDATA, op->pid, from, 0);
		ucp = (unsigned char *) &word;
		for (i = 0; i < sizeof(word); i++) {
			*(dest + nbytes++) = *(ucp + i);
			if (nbytes >= len)
				break;
		}
		if (nbytes >= len)
			break;
	}

	return (nbytes);
}

/*
 * Write the memory block into debuggee address space
 */
int
xtr_write_mem(const struct opt *op, vaddr_t to, const void *from, int len)
{
	unsigned long *p;
	int n, reallen;
	
	/* allocate dword-aligned block */
	reallen = len / sizeof(long) + sizeof(long);
	if ((p = malloc(reallen)) == NULL) {
		dprintf((stderr, "*** xtr_write_mem: malloc failed\n"));
		return (-1);
	}

	/* read memory inside that block */
	(void) xtr_read_mem(op, to, p, reallen);

	/* now copy the data into this block */
	(void) memcpy(p, from, len);

	/* write it back to debuggee's address space */
	for (n = 0; n < reallen / sizeof(long); n++)
		if (ptrace(PTRACE_POKEDATA, op->pid, to, *p++) == -1) {
			dprintf((stderr, "*** xtr_write_mem: ptrace: %s\n",
			     strerror(errno)));
			break;
		}

	return (n * sizeof(long));
}
