/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   System V RELEASE 4 - specific code.
 *   this file is a part of xtrace project
 *
 * $Id: svr4.c,v 1.4 2002/04/21 08:11:31 drozd Exp $
 * $Log: svr4.c,v $
 * Revision 1.4  2002/04/21 08:11:31  drozd
 * autoconf support. svr4 is still broken
 *
 * Revision 1.3  2002/02/24 18:25:29  drozd
 * change in interface - svr4 stuff is broken now
 *
 * Revision 1.2  2002/02/21 13:49:56  drozd
 * function trace enabled
 *
 * Revision 1.1.1.1  2002/02/21 08:55:10  drozd
 * import
 *
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/syscall.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <dirent.h>
#include <fcntl.h>
#include <ctype.h>
#include <procfs.h>

#include "xtrace.h"
#include "svr4.h"


static struct svr4data __data;

/* syscalls array */
#include "svr4-syscalls.c"

static void makeset(sysset_t *, char *);
static void finiwatch(struct opt *);
static void initwatch(int, char *);
static void dumplwp(pid_t, id_t);
static void dumpcall(struct opt *);
static void fini(struct opt *);

/*
 * Return list of ELF filenames that executable references to
 * Parameters:
 *   pid - PID of a process
 * Return:
 *   NULL - terminated array of pointers to filenames
 */
char *const *
xtr_get_elfs(pid_t pid)
{
	static char * fnames[128];	/* maximum - 128 ELF files */
	int i;
	DIR *dirp = NULL;
	struct dirent *dep = NULL;
	char dirname[128], fname[128];

	/* open /proc/<pid>/object directory */
	(void) snprintf(dirname, sizeof(dirname), "/proc/%d/object", pid);
	if ((dirp = opendir(dirname)) == (DIR*) NULL) {
		perror("filltab opendir()");
		return 0;
	}

	/* for each entry in object directory  */
	for (i = 0; (dep = readdir(dirp)) != (struct dirent *) 0; ) {

		if (!strcmp(dep->d_name, ".") || !strcmp(dep->d_name, ".."))
			continue;

		(void) snprintf(fname, sizeof(fname),"%s/%s", dirname,
		    dep->d_name);

#ifdef __DEBUG__
		(void) fprintf(stderr, "filltab: trying %s\n", fname);
#endif /* __DEBUG__ */

		if (i >= sizeof(fnames)/sizeof(fnames[0])) {
			(void) fprintf(stderr, "Warning: too many mappings\n");
			break;
		}

		/* add filename to a table */
		fnames[i++] = strdup(fname);
	}

	return fnames;
}

/*
 * Initialize tracer
 * Parameters:
 *    op - current options
 *    slist - comma separated list of syscalls to trace
 *    flist - comma separated list of functions to trace
 * Return:
 *    0 on success, < 0 on error
 */
int
xtr_init(struct opt *op, char *slist, char *flist)
{
	int fd;
	struct pctl pctl;
	struct stat st;
	struct pstatus ps;
	char fname[32];
	long cmd;
	struct svr4data *dp;

	/* register platform specific cleanup function and data */
	op->fini = fini;
	op->data = dp = &__data;
	op->scall = op->fcall = dumpcall;

	/* find the process in the procfs */
	(void) snprintf(fname, sizeof(fname), "%s%d", "/proc/", op->pid);

	/* accessible ? */
	if (access(fname, X_OK) < 0) {
		(void) fprintf(stderr, "%s\n", fname);
		perror("access()"); return -1;
	}

	/* open control file, status and address space file */
	(void) snprintf(fname, sizeof(fname), "/proc/%d/ctl", op->pid);
	if ((dp->cfd = open(fname, O_WRONLY )) < 0) {
		perror("ctl open()"); return -1;
	}
	(void) snprintf(fname, sizeof(fname), "/proc/%d/status", op->pid);
	if ((dp->sfd = open(fname, O_RDONLY)) < 0) {
		perror("status open()"); return -1;
	}
	(void) snprintf(fname, sizeof(fname), "/proc/%d/as", op->pid);
	if ((dp->afd = open(fname, O_RDONLY)) < 0) {
		perror("address space open()"); return -1;
	}

	/* look - is it a multithreaded program */
	if (read(dp->sfd, &ps, sizeof(ps)) != sizeof(ps)) {
		perror("status read()"); return -1;
	}
	if ((op->flags & XTR_TRACEFUNC) && (ps.pr_nlwp > 1)) {
		(void) fprintf(stderr, "Cannot do function trace on"
		    " multithreaded application\n");
		return -1;
	}

	/* stop the debuggee */
	cmd = PCSTOP;
	if (write(dp->cfd, &cmd, sizeof(cmd)) != sizeof(cmd)) {
		perror("write(PCSTOP)"); return -1;
	}

	/* open map file, filling up global pmap structure */
	(void) snprintf(fname, sizeof(fname), "/proc/%d/map", op->pid);
	if ((fd = open(fname, O_RDONLY)) == -1) {
		perror("map open()"); return -1;
	}
	if (fstat(fd, &st) == -1) {
		perror("map fstat()"); return -1;
	}

	/* set up __pmap */
	dp->nregions = (unsigned long) st.st_size/sizeof(prmap_t) + 1;
	dp->prmap = (prmap_t *) calloc(dp->nregions, sizeof(prmap_t));

	if (read(fd, (void *)dp->prmap, (size_t)st.st_size) !=
	    (int) st.st_size) {
		perror("procfs_map map read()");
		free(dp->prmap);
		dp->prmap = 0;
	}

	(void) close(fd);

	/* allow debuggee to run when we detach */
	pctl.cmd = PCSET;
	pctl.data = PR_RLC | PR_FORK;
	if (write(dp->cfd, &pctl, sizeof(struct pctl)) != sizeof(struct pctl)) {
		perror("write(PCSET PR_RLC)"); return -1;
	}

	/* install watchpoints */
	if (op->flags & XTR_TRACEFUNC)
		initwatch(dp->cfd, flist);

	/* trace system calls */
	if (op->flags & XTR_TRACESYS) {
		dp->sctl.cmd = PCSEXIT;
		makeset(&dp->sctl.sysset, slist);
		if (write(dp->cfd, &dp->sctl, sizeof(struct sctl)) !=
		    sizeof(struct sctl)) {
			perror("write(PCSEXIT)"); return -1;
		}
	}
	
	/* resume process */
	pctl.cmd = PCRUN;
	pctl.data = PRCFAULT | PRSABORT;
	if (write(dp->cfd, &pctl, sizeof(struct pctl)) != sizeof(struct pctl)) {
		perror("write(PCRUN)");
	}

	return 0;
}

/*
 * Prepare syscall set to trace
 * Parameters:
 *   sp - syscall set ptr
 *   argp - comma-separated character string
 * Return:
 *   void
 */
void
makeset(sysset_t *sp, char *argp)
{
	char *commap;
	int found, i;

	premptyset(sp);

	/* trace all syscalls */
	if (!strcmp(argp, "all")) {
		prfillset(sp);
		return;
	}

	/* scan argument string, adding syscalls to the set */
	for (commap = argp; argp != 0; ) {

		commap = strchr(argp, ',');

		if (commap != 0)
			*commap++ = '\0';
		else
			commap = 0;

		found = -1;
		for (i = 0; i < sizeof(__syscalls)/sizeof(struct syscall); i++)
			if (!strcmp(argp, __syscalls[i].name)) {
				++found;
				praddset(sp, __syscalls[i].no);
				break;
			}
		if (found != 0) {
			(void) fprintf(stderr, "unknown syscall %s\n", argp);
			exit(1);
		}
			
		argp = commap;
	}
}

/*
 * Set the watchpoints on function prologues
 * XXX: debuggee must be stopped when calling this function
 * Parameters:
 *   fd - ctl file descriptor
 * Return:
 *   void
 */
void
initwatch(int fd, char *flist)
{
	struct wctl wctl;
	struct fctl fctl;
	int doall, found, nbytes;
	char *commap, *argp;
	struct symbol **symp;

	/* are we initialized OK ? */
	if (fd <= 0)
		return;		/* no */


	/* catch watch faults */
	fctl.cmd = PCSFAULT;
	premptyset(&fctl.fltset);
	praddset(&fctl.fltset, FLTWATCH);
	if (write(fd, &fctl, sizeof(fctl)) != sizeof(fctl)) {
		perror("write(PCSFAULT)");
		return;
	}

	/* we are watching functions, so trap on EXEC */
	wctl.cmd = PCWATCH;
	wctl.w.pr_wflags = WA_EXEC | WA_TRAPAFTER;
	/* length of one instruction or whole address space */
	wctl.w.pr_size = 4;

	if (!strcmp(flist, "all"))
		doall = 1;
	else
		doall = 0;


	 for (symp = 0; (symp = xtr_get_next_sym(symp)) != 0; ) {

		switch (doall) {

			case 1:
				found = -1;

				if (!((*symp)->flags & SYM_SETWATCH))
					continue;

				++found;
				break;

			case 0:
				/* scan argument list  */
				for (commap = argp = flist, found = -1;
				    argp != 0; argp = commap) {

					commap = strchr(argp, ',');

					if (commap != 0)
						commap++;
					else
						commap = 0;

					nbytes = commap == 0
						? strlen(argp)
						: commap - argp - 1;
					if (strncmp((*symp)->name, argp, nbytes))
						continue;

					if ((*symp)->name[nbytes] == '\0')
						++found;
				}
				break;
				
			default:
				found = -1;
				break;
		}

		if (found != 0)
			continue;

		/* set the watchpoint */
		wctl.w.pr_vaddr = (*symp)->va;
		if (write(fd, &wctl, sizeof(wctl)) != sizeof(wctl)) {
			perror("write(PCWATCH)");
			(void) fprintf(stderr,
			    "Warning: not all watchpoints were installed!\n");
		}
	}
}

/*
 * Exit handler
 */
void
fini(struct opt *op)
{
	struct svr4data *dp = (struct svr4data *) op->data;

	finiwatch(op);

	/* close /proc file descriptors */
	(void) close(dp->cfd);
	(void) close(dp->afd);
	(void) close(dp->sfd);
}

/*
 * Set the watchpoints on function prologues
 * XXX: debuggee must be stopped when calling this function
 * Parameters:
 *   fd - ctl file descriptor
 * Return:
 *   void
 */
void
finiwatch(struct opt *op)
{
	struct wctl wctl;
	struct svr4data *dp = (struct svr4data *) op->data;

	/* if no watchpoints were set, quit */
	if (!(op->flags & XTR_TRACEFUNC))
		return;

	wctl.cmd = PCWATCH;
	wctl.w.pr_wflags = 0;		/* unset watches */
	wctl.w.pr_size = 0xffffffff;	/* in whole address space */
	wctl.w.pr_vaddr = 0U;		/* starting from 0 */

	if (write(dp->cfd, &wctl, sizeof(wctl)) != sizeof(wctl)) {
		perror("write(PCWATCH)");
		(void) fprintf(stderr, "error removing watchpoints");
	}               
}

/* 
 * Start the application, ordering to stop after execve()
 * Parameters:
 *    argv - argument vector
 *    envp - environment
 * Return:
 *    PID of new process
 */
pid_t
xtr_start_app(char *const *argv, char *const *envp)
{
	pid_t pid;
	char str[128];
	int cfd;
	struct sctl sctl;

	pid = fork();

	if (pid != (pid_t) 0)
		return pid;


	/* order to stop after exec() */
	(void) snprintf(str, sizeof(str), "/proc/%d/ctl", (int) getpid());
	if ((cfd = open(str, O_WRONLY)) < 0) {
		perror("kid's open(ctl)"); return -1;
	}

	premptyset(&sctl.sysset);
	prfillset(&sctl.sysset);
	sctl.cmd = PCSEXIT;

	if (write(cfd, &sctl, sizeof(sctl)) != sizeof(sctl)) {
		perror("kid's write(ctl) "); return -1;
	}

	(void) execve(argv[0], argv, envp);

	/* NOTREACHED */
	return 0;
}

/*
 * Stop on event of interest, is a blocking function
 * Parameters:
 *    op - ptr to the options
 * Return:
 *    void
 */
int
xtr_stop_app(const struct opt *op)
{
	long cmd = PCWSTOP;
	int retval = 0;
	struct svr4data *dp = (struct svr4data *) op->data;

	
	if (write(dp->cfd, &cmd, sizeof(cmd)) != sizeof(cmd)) {
			perror("write(PCTWSTOP)");
			--retval;
	}

	return retval;
}

/*
 * Resume the appliction
 */
int
xtr_resume_app(const struct opt *op)
{
	struct pctl pctl;
	int retval = 0;
	struct svr4data *dp = (struct svr4data *) op->data;

	pctl.cmd = PCRUN;
	pctl.data = PRCFAULT | PRSABORT;
	if (write(dp->cfd, &pctl, sizeof(pctl)) != sizeof(struct pctl)) {
		perror("write(PCRUN)");
		--retval;
	}
	
	return retval;	
}

/*
 * Dump memory location
 * Parameters:
 *   op - ptr to options
 *   va - virtual address to start from
 *   len - length to dump
 * Return: void
 */
void
xtr_dump_mem(const struct opt *op, vaddr_t va, int len)
{
	char buf[512], out[BUFSIZ] = "(\"";
	int i, j, dumpsiz;
	struct svr4data *dp = (struct svr4data *) op->data;
	
	
	dumpsiz = len < sizeof(buf) ? len : sizeof(buf);

	/* first page is always not mapped to prevent NULL ptrs dereferences */
	if (va < 0x1000U || dp->prmap == 0)
		return;
	
	/* looping over mapped region removed, now try to read blindly */

	if (lseek(dp->afd, va, SEEK_SET) == (off_t) -1)
		return;
	if (read(dp->afd, buf, dumpsiz) != dumpsiz)
		return;

	/* i is the counter for buf, j - for out */
	for (i = 0, j = 2; i < dumpsiz && j < sizeof(out) - 3; i++) {
		if (isprint((int)buf[i]))
			out[j++] = buf[i];
		else {
			out[j++] = '\\';
			j += sprintf(out + j, "%x", buf[i]);
		}
	}
	out[j++] = '"', out[j++] = ')', out[j++] = '\0';
	(void) printf("%s", out);
}

/*
 * Dump information about current call
 * Parameters:
 *   sp - ptr to struct prstatus, should be read from /proc/<pid>/status
 * Return:
 *   void
 */
void
dumpcall(struct opt *op)
{
	struct pstatus ps;
	char tim[32] = "";
	vaddr_t *args;
	vaddr_t pc;		/* instruction pointer */
	ptrdiff_t offt;		/* offset from prologue */
	int ncol;		/* number of character printf-ed on terminal */
	int nbytes, i;
	struct lwpstatus *lwpp = &ps.pr_lwp;
	struct svr4data *dp = (struct svr4data *) op->data;
	const struct syscall *scp;
	const struct symbol *symp;

	/* set timestamp if flag set */
	if (op->flags & XTR_TIMESTAMP)
		(void) snprintf(tim, sizeof(tim), "(%lu.%lu)",
		    lwpp->pr_tstamp.tv_sec, lwpp->pr_tstamp.tv_nsec);


	
	(void) lseek(dp->sfd, 0, SEEK_SET);
	if ((nbytes = read(dp->sfd, &ps, sizeof(ps))) != sizeof(ps)) {
		perror("read(/proc/pid/status)");
		exit(1);
	}

	/* dump per-thread stuff if requested */
	if ((op->flags & XTR_DUMPTHREADS) && (
	    (((op->flags & XTR_TRACEFUNC) && lwpp->pr_why == PR_FAULTED)
	    || ((op->flags & XTR_TRACESYS) && lwpp->pr_why == PR_SYSEXIT))))
		for (i = 1; i <= ps.pr_nlwp; i++)
			dumplwp(op->pid, (id_t) i);

	switch (lwpp->pr_why) {

		case PR_SYSEXIT:

			/*
			 * don't dump syscalls if function trace is switched on
			 * and __tflag is not set
			 */	
			if (!(op->flags & XTR_TRACESYS))
				break;

			scp = xtr_lookup_syscall(lwpp->pr_syscall);
			ncol = printf("%lu <%s>(", lwpp->pr_lwpid, scp->name);

			for (i = 0; i < lwpp->pr_nsysarg; i++) {
				ncol += printf("%lx", lwpp->pr_sysarg[i]);
				/* if valid address, dump the content */
				if (op->flags & XTR_DUMPMEM)
					xtr_dump_mem(op,
					    lwpp->pr_sysarg[i], 0x10);
				if (i != lwpp->pr_nsysarg - 1) {
					(void) putchar(',');
					++ncol;
				}
			}

			(void) putchar(')');

			/* align output */
			for (; ncol <= 40; ncol++)
				(void) putchar(' ');

			/* dump return value */
			(void) printf("= %-8lx %s\n", lwpp->pr_rval1, tim);
			break;

		case PR_FAULTED:

			/* we're in watchpoint trap */
			pc = lwpp->pr_reg[R_PC];
			symp = xtr_find_symbol(pc, &offt);
			(void) printf("%lu [%s](", lwpp->pr_lwpid, symp->name);
			args = (vaddr_t *) &lwpp->pr_reg[R_I0];
			for (i = 0; i < 6; i++) {
				(void) printf("%x", args[i]);
				if (op->flags & XTR_DUMPMEM)
					xtr_dump_mem(op, args[i], 0x10);
				if (i != 5)
					(void) putchar(',');
			}
			(void) printf(") + %x, pc: %x %s\n", offt, pc, tim);

			break;

		default:
			(void) fprintf(stderr, "unexpected stop reason %d\n",
			     lwpp->pr_why);
	}
}

/*
 * Dump lwp information and call stack.
 * Parameters:
 *   pid  - a PID of a process to be examined
 * Return:
 *   void
 */
void
dumplwp(pid_t pid, id_t lwpid)
{
	int fd;				/* procfs file descriptor */
	char fname[64];			/* procfs file name */
	struct lwpstatus st;		/* lwp status struc */
	/*
	const char *symp;
	unsigned *args;
	unsigned pc;
	ptrdiff_t diff;
	*/

	(void) printf("LWP %d ===========================\n", (int) lwpid);

	/* open status file */
	(void) snprintf(fname, sizeof(fname),
	    "/proc/%d/lwp/%d/lwpstatus", pid, lwpid);

	if ((fd = open(fname, O_RDONLY)) == -1) {
		perror("procfs_lwp lwpstatus open()");
		return;
	}

	if (read(fd, &st, sizeof(st)) != (int) sizeof(st)) {
		perror("procfs_lwp lwpstatus read()");
		close(fd);
		return;
	}

	(void) printf(
	    "flags: %#-10x  lwpid: %-10lu  why: %-10i     what: %-10i\n"
	    "cursig: %-10i oldctx: %#-10x syscall: %-10i nargs: %-10i\n"
	    "errno: %-10d  PC: %#-10x\n",
	    st.pr_flags, st.pr_lwpid, st.pr_why, st.pr_what,
	    st.pr_cursig, st.pr_oldcontext, st.pr_syscall, st.pr_nsysarg,
	    st.pr_errno, st.pr_reg[R_PC]);

	(void) printf("Call stack:\n");

	/* close file descriptor */
	(void) close(fd);
}

/*
 *
 */
void
xtr_set_exec_callback(struct opt *op)
{
	struct sctl sctl;
	int fd;
	int flags;
	char fname[64];

	UNUSED(op);

	(void) snprintf(fname, sizeof(fname), "/proc/%d/ctl", getpid());
	fd = open(fname, O_WRONLY);

	/* set close-on-exec */
	flags = fcntl(fd, F_GETFD);
	flags |= FD_CLOEXEC;
	(void) fcntl(fd, F_SETFD, flags);

	premptyset(&sctl.sysset);
	praddset(&sctl.sysset, SYS_execve);
	sctl.cmd = PCSEXIT;

	(void) write(fd, &sctl, sizeof(struct sctl));
}

/*
 *
 */
vaddr_t
xtr_get_load_addr(const struct opt *op, const char *elfname)
{
	vaddr_t retval = 0;
	struct svr4data *dp = (struct svr4data *) op->data;
	int i;

	if (dp->prmap == 0)
		return 0;

	for (i = 0; i < dp->nregions; i++)
		if (strcmp(dp->prmap[i].pr_mapname, elfname)) {
			retval = dp->prmap[i].pr_vaddr;
			break;
		}

	return retval;
}
