/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   Symbol table manipulation routines
 *   this file is a part of xtrace project
 *
 * $Id: symbol.c,v 1.12 2002/09/23 16:51:17 drozd Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

#include "xtrace.h"
#include "elf.h"

/* symbol table - array of pointers to struct symbol */
static struct symbol **__symtab;

static int __nsyms = -1;	/* number of symbols in symbol table */


/* local helpers */
static void handlesym(struct opt *, Elf32_Ehdr *,
		Elf32_Shdr *, Elf32_Sym *, vaddr_t);
static int sortsyms(const void *, const void *);
static void addsym(struct opt *, const char *, vaddr_t, unsigned);

/*
 * Fill up the debuggee symbol table. The only symbols that are referenced from
 * debugged executable are used. For every such symbol, the only ones that are
 * `defined' in ELF object, are used. Say, debuggee depends on libc.so, and
 * imports `strcmp' from there, and depends on libsocket.so as well. There
 * will be at least two references to `strcmp' - one from debuggee, and one
 * from libsocket, and both marked as `undefined'. 
 * The code will add symbol for `strcmp' symbol just for libc.so mapping,
 * 'cause `strcmp' is defined in .text section only in libc.so ELF mapping.
 * Parameters:
 *   op - ptr to the options
 * Return:  void
 */
int
xtr_symtab_build(struct opt *op)
{
	struct sharedlib **deps;
	int i, k;
	int nentries;			/* number of entries in symbol sec */
	int fd;				/* fd for mapped object */
	Elf32_Ehdr *ep;
	Elf32_Shdr *shdrp;
	Elf32_Sym *symp;
	
	dprintf((stderr, "*** xtr_symtab_build\n"));

	/* for each ELF */
	for (deps = xtr_get_deps(op->pid); deps != 0 && *deps != 0; deps++) {

		if ((fd = open((*deps)->path, O_RDONLY)) < 0) {
			dprintf((stderr, "*** xtr_symtab_build:"
			    "%s: %s", (*deps)->path, strerror(errno)));
			free((*deps)->path);
			continue;
		}

		if ((ep = elf_init(fd)) == 0) {
			dprintf((stderr, "xtr_symtab_build: "
			    "elf_init() failed\n"));
			(void) close(fd);
			continue;
		}

		dprintf((stderr, "*** xtr_symtab_build: reading symbols"
		    " for %s (%x)\n", (*deps)->path, (*deps)->base));
		free((*deps)->path);

		/* find the symbol table section */
		for (i = 0; i < ep->e_shnum; i++ ) {
			shdrp = elf_getshdr(ep, i);

			if (!(shdrp->sh_type == SHT_SYMTAB ||
			    shdrp->sh_type == SHT_DYNSYM))
				continue;

			symp = (Elf32_Sym *) ((char *) ep + shdrp->sh_offset);

			/* scan all symbols in symbol section  */
			nentries = shdrp->sh_size/shdrp->sh_entsize;
			for (k = 0; k < nentries; k++, symp++) {
				handlesym(op, ep, shdrp, symp, (*deps)->base);
			}
		}

		/* close file */
		(void) elf_fini(ep);
		(void) close(fd);
	}

	/* now sort symbols */
	if (__nsyms > 0)
		qsort(__symtab, __nsyms + 1, sizeof(struct symbol *), sortsyms);

	/* dump function list and exit if requested */
	if (op->flags & XTR_FUNCLIST) {
		for (k = 0; k < __nsyms; k++)
			if ((__symtab[k])->flags & SYM_REFERENCED)
				(void) printf("%-8x %s\n", (__symtab[k])->va,
				     (__symtab[k])->name);
		UNLOCK(op);
		exit(0);
	}

	return (__nsyms >= 0 ? 0 : -1);
}

/*
 * Helper function for xtr_symtab_build(). must be called for
 * every entry in ELF object's symbol table.
 * Parameters:
 *   ep - ELF object descriptor
 *   shdrp - section header ptr
 *   symp - ptr to symbol table entry
 *   base - load address for the  given ELF object
 * Return:
 *   void
 */
void
handlesym(struct opt *op, Elf32_Ehdr *ep, Elf32_Shdr *shdrp,
	Elf32_Sym *symp, vaddr_t base)
{
	Elf32_Shdr *thdrp;
	unsigned flags;
	const char *symname;		/* symbol name */
	vaddr_t va;			/* virtual address, == rva + base */

	if (ep == 0)
		return;

	/* beginning of section table */
	
	flags = base == 0U ? SYM_REFERENCED : SYM_SETWATCH;

	if (ELF32_ST_TYPE(symp->st_info) != STT_FUNC)
		return;

	symname = elf_str(ep, shdrp->sh_link, symp->st_name);
	va = symp->st_value + base;

	/* for a.out, allow watch on symbols from .text */
	if (base == 0) {
		if (symp->st_shndx == elf_getindex(ep, ".text", &thdrp))
			flags |= SYM_SETWATCH;
	}

	addsym(op, symname, va, flags);
}

/*
 * Sort symbols by virtual addresses, to be passed to qsort()
 */
int
sortsyms(const void *p1, const void *p2)
{
	return ((* (struct symbol **) p1)->va >
			(* (struct symbol **) p2)->va ? 1 : -1);
}

/*
 * Add a symbol in a symbol table
 * Parameters:
 *   name - symbol's name
 *   va - virtual address in debuggee address space
 *   flags - symbol's flags
 * Return:
 *   element's index in a table
 */
void
addsym(struct opt *op, const char *name, vaddr_t va, unsigned flags)
{
	static int tsiz;		/* size of symbol table */
	struct symbol *symp;
	int i;
	/* allocate space for 65K bytes in one go */
#define BUCKETSIZ				0x10000

	/* look for duplication in virtual addresses */
	for (i = 0; i < __nsyms; i ++)
		if (__symtab[i]->va == va) {
#if 0
			dprintf((stderr, "*** addsym: duplicate va:"
			   "%s,%s (%x)\n", name, __symtab[i]->name, va));
#endif
			return;
		}

	/* first allocation  */
	if (__symtab == (struct symbol **) 0) {
		if ((__symtab=calloc(BUCKETSIZ,sizeof(struct symbol *))) == 0){
			perror("symtab calloc()");
			UNLOCK(op);
			exit(1);
		}
		tsiz = BUCKETSIZ / sizeof(struct symbol *);
	}

	/* allocate symbol */
	if ((symp = (struct symbol *) calloc(1, sizeof(struct symbol))) == 0) {
		perror("addsym calloc()");
		UNLOCK(op);
		exit(1);
	}
	symp->name = strdup(name);
	symp->va = va;
	symp->flags = flags;

	/* extend the table if need */
	if (tsiz <= __nsyms) {
		tsiz += BUCKETSIZ / sizeof(struct symbol *);
		if ((__symtab = (struct symbol **) realloc(__symtab, tsiz *
		    sizeof(struct symbol *))) == (struct symbol **) 0) {
			perror("addsym realloc()");
			UNLOCK(op);
			exit(1);
		}
		dprintf((stderr, "*** __symtab realloced\n"));
	}
	dprintf((stderr, "*** addsym: %x %s\n", va, name));

	/* increment global symbol counter and install the symbol */
	__symtab[++__nsyms] = symp;
}

/*
 * Clear symbol table
 * Parameters:
 * Return:
 * void
 */
void
xtr_symtab_destroy(void)
{
	dprintf((stderr, "*** xtr_symtab_destroy: %d syms\n", __nsyms));
	if (__nsyms < 0)
		return;

	for (; __nsyms >= 0; __nsyms--) {
		free(__symtab[__nsyms]->name);
		free(__symtab[__nsyms]);
	}

	free(__symtab);
	__symtab = (struct symbol **) 0;
	__nsyms = -1;
}

/*
 * Iterate over symbols
 * Parameters:
 * Return:
 *    ptr to the next symbol, or NULL if end or no symbols
 */
struct symbol *
xtr_get_next_sym(void)
{
	static int idx;

	if (__nsyms <= 0)	/* no symbols */
		return (NULL);

	if (idx > __nsyms) {
		idx = 0;
		return (0);
	}

	return (__symtab[idx++]);
}

/*
 * Return symbol name from given vaddr in ELF file, symbol name is the
 * function's name which address is closest to given.
 * 3rd agrument returns the difference
 * Parameters:
 *   addr - address inside the function body
 *   where_called - will contain the offset between function start and addr
 * Return:
 *   symbol name
 */
struct symbol *
xtr_find_symbol(vaddr_t addr, ptrdiff_t *where_called)
{
	int idx;

	if (__nsyms <= 0)
		return 0;

	for (idx = 0; idx <= __nsyms; idx++)
		if (__symtab[idx]->va > addr)
			break;

	if (idx > 0)
		idx--;

	if (where_called != 0)
		*where_called = addr - __symtab[idx]->va;

	dprintf((stderr, "*** xtr_find_symbol: va %x, offset %x\n",
				addr, where_called ? *where_called : 0));

	return __symtab[idx];
}

