/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   Main module for xtrace utility
 *   this file is a part of xtrace project
 *
 * $Id: xtrace.c,v 1.18 2002/09/25 10:49:42 drozd Exp $
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include "xtrace.h"

static const char __version[] =
"$Id: xtrace.c,v 1.18 2002/09/25 10:49:42 drozd Exp $";

/* options */
static struct opt __opt;

static void prepare(struct opt *);
static void start_app(struct opt *, char *const *, char *const *);
static void fini(int);
static void chldhndlr(int);
static int parseargs(int argc, char *const *argv, struct opt *op);
static char *checkopt(char *const *argv, int *optind);

/*
 * Print usage information and exit
 * Parameters:
 *   pname - executable's name
 * Return value:
 *   void
 */
void usage(const char *pname)
{
	(void) fprintf(stderr, "usage:\n%s [OPTIONS] [filename ...]\n"
	    "OPTIONS:\n"
	    "-p <pid>\tPID of a process to be traced\n"
	    "-h\t\tthis help\n"
	    "-V\t\tprint version and exit\n"
	    "-S\t\tdump stack trace\n"
	    "-T\t\tdump per-thread info (unimplemented)\n"
	    "-o <file>\tredirect output to file\n"
	    "-t\t\tput timestamp for every syscall\n"
	    "-d <siz>\tshow memory content of <siz> bytes for valid addresses\n"
	    "-s <syscall[,syscall2...]>\ttrace specific syscalls\n"
	    "-f <func1,[func2...]|all>\ttrace functions\n"
	    "-L\t\tdump list of functions from executable and exit\n"
	    "Note: -v, -S, -f options affect perfomance seriously!\n",
	    pname);
	exit(1);
}

/*
 * Option sanity check
 * Return the ptr to option argument
 */
char *
checkopt(char *const *argv, int *optind)
{
	char *retval = NULL;

	switch (argv[*optind][1]) {
		case 'p':
		case 'f':
		case 's':
		case 'd':
		case 'o':
			if (argv[*optind][2] == '\0')
				retval = argv[++(*optind)];
			else
				retval = &argv[*optind][2];
			break;
		default:
			if (argv[*optind][2] != '\0')
				usage(argv[0]);
			break;
	}

	return (retval);
}


/*
 * Parse command line parameters
 */
int
parseargs(int argc, char *const *argv, struct opt *op)
{
	int i;

	for (i = 1; i < argc; i++) {
		
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'p':
					op->pid = (pid_t) atoi(checkopt(argv, &i));
					break;
				case 't':
					(void) checkopt(argv, &i);
					op->flags |= XTR_TIMESTAMP;
					break;
				case 's':
					op->flags |= XTR_TRACESYS;
					op->slist = checkopt(argv, &i);
					break;
				case 'o':
					(void) freopen(checkopt(argv, &i), "w", stdout);
					break;
				case 'd':
					op->memdumpsiz = atoi(checkopt(argv, &i));
					op->flags |= XTR_DUMPMEM;
					break;
				case 'f':
					op->flags |= XTR_TRACEFUNC ;
					op->flist = checkopt(argv, &i);
					break;
				case 'S':
					(void) checkopt(argv, &i);
					op->flags |= XTR_BACKTRACE;
					break;
				case 'T':
					(void) checkopt(argv, &i);
					op->flags |= XTR_DUMPTHREADS;
					break;
				case 'L':
					(void) checkopt(argv, &i);
					op->flags |= XTR_FUNCLIST;
					break;
				case 'V':
					(void) checkopt(argv, &i);
					(void) printf("%s\n", __version);
					exit(0);
					break;
				case 'h':
					/* FALLTHROUGH */
				default:
					usage(argv[0]);
					/* NOTREACHED */
					break;

			}
		} else {
			break;		/* no options left */
		}
	}

	return (i);
}

/*
 * jump point for crt1.o code
 */
int
main(int argc, char *const *argv, char *const *envp)
{
	volatile int retval, optind;
	struct opt *op = &__opt;

	optind = parseargs(argc, argv, op);

	/*
	 * clear the lock immediately - it is closed by default (0)
	 * so fini() will not hang
	 */
	UNLOCK(op);

	if (argc == 1)
		usage(argv[0]);

	/* check child's PID sanity and start the child if requested */
	if (argc != optind)
		start_app(op, &argv[optind], envp);
	else if (op->pid <= (pid_t) 0)
		usage(argv[0]);

	prepare(op);

	/* infinite trace loop */
	for (retval = -1; retval != 0; ) {

		/* stop on event of interest */
		(void) xtr_stop_app(op);
					
		/* rebuild the symbol table after execve() */
		if (BIT_ISSET(op->flags, XTR_INEXEC)) {
			if (SYMTAB_NEEDED(op)) {
				xtr_symtab_destroy();
				xtr_symtab_build(op);
				if (BIT_ISSET(op->flags, XTR_TRACEFUNC))
					xtr_set_func_breakpoints(op);
			}

			BIT_UNSET(op->flags, XTR_INEXEC);
		}

		/* show it */
		xtr_dump_call(op);

		/* exit, if child is dead */
		if (BIT_ISSET(op->flags, XTR_CHILDEXITED)) {
			UNLOCK(op);
			exit(0);
		}

		/* resume process */
		(void) xtr_resume_app(op);
	}

	return 0;
}

/*
 * atexit handler
 * @signo - signal number caught
 * Return: void
 */
void
fini(int signo)
{
	struct opt *op = &__opt;

	LOCK(op);

	/* unset the breakpoints */
	xtr_clrbpts(op);
	
	/* call platform specific cleanup */
	if (op->fini != 0)
		(*op->fini)(op);

	xtr_symtab_destroy();

	(void) fflush(stderr);
	(void) fflush(stdout);

	_exit(0);
}

/*
 * SIGCHLD handler
 */
void
chldhndlr(int signo)
{
	struct opt *op = &__opt;
	const char *reason;
	int status;

	UNLOCK(op);

	(void) wait(&status);
	op->signo = WSTOPSIG(status);

	if (WIFSIGNALED(status))
		reason = "signalled";
	else if (WIFEXITED(status))
		reason = "exited";
	else if (WIFSTOPPED(status))
		reason = "stopped";
	else
		reason = "unknown";
	
	dprintf((stderr, "*** chldhndlr: [%s] (sig %d, flags %#x %s)\n",
	    reason, op->signo, op->flags,
	    op->flags & XTR_INFENTER ? "fenter" : "nofenter"));

	/* call platform-specific handler */
	if (op->child != NULL)
		(*op->child)(op);

}


/*
 * Platform-independet xtr_init()
 * Parameters:
 *    op - ptr to options
 *    slist - comma separated list of syscalls to trace
 *    flist - comma separated list of functions to trace
 * Return: void
 */
void
prepare(struct opt *op)
{
	/* trace syscalls by default */
	if (!(op->flags & XTR_TRACESYS) && !(op->flags & XTR_TRACEFUNC)) {
		op->flags |= XTR_TRACESYS;
		op->slist = "all";
	}

	(void) atexit((void (*)(void)) fini);

	(void) signal(SIGINT, fini);
	(void) signal(SIGTERM, fini);

	/* call platform dependent init function */
	if (xtr_init(op) < 0) {
		(void) fprintf(stderr, "error init xtrace\n");
		exit(1);
	}

	/* set SIGCLD handler to fini if xtr_init() didn't set it */
	(void) signal(SIGCHLD, chldhndlr);

	/* fill up symbol table */
	if (SYMTAB_NEEDED(op))
		(void) xtr_symtab_build(op);

	if (BIT_ISSET(op->flags, XTR_TRACEFUNC))
		xtr_set_func_breakpoints(op);

	/* mark as attached */
	op->flags |= XTR_ATTACHED;
}

/*
 * Set the breakpoints on function prologues
 * Parameters:
 *   op - ptr to options
 *   flist - comma separated list of functions to break
 * Return: void
 */
void
xtr_set_func_breakpoints(struct opt *op)
{
	struct symbol *symp;
	char *ptrs[1024];		/* ptrs to function names */
	char *curr, *next, *flist;
	int i, set = 0;

	flist = strdup(op->flist);

	/*
	 * prepare ptrs. change each ',' to '\0' and put the pointer value
	 * into the array.
	 */
	for (curr = flist, i = 0; (next = strchr(curr, ',')) != NULL; i++) {
		ptrs[i] = curr;
		curr[next - curr] = '\0';
		curr = next + 1;
	}
	ptrs[i] = curr;
	ptrs[++i] = NULL;

	/* set bpt on every function if "-fall" flag given */
	if (!strcmp("all", op->flist)) {
		for (symp = NULL; (symp = xtr_get_next_sym()) != NULL; ) {
			xtr_set_bpt(op, symp->va);
			++set;
			dprintf((stderr, "*** bpt set on %s (%#x)\n",
			    symp->name, symp->va));
		}
	}

	/* if not, loop over each function name */
	for (symp = NULL; (symp = xtr_get_next_sym()) != NULL; ) {
		for (i = 0; ptrs[i] != NULL; i++) {
			if (!strcmp(ptrs[i], symp->name)) {
				xtr_set_bpt(op, symp->va);
				++set;
				dprintf((stderr, "*** bpt set on %s (%#x)\n",
				    symp->name, symp->va));
			}
		}
	}

	free(flist);

	if (!set)
		(void) fprintf(stderr, "set_func_bpts: no breakpoints set\n");
}

/* 
 * Start the application, ordering to stop after execve()
 * Parameters:
 *    argv - arguments vector
 *    envp - environment
 * Return: void
 */
void
start_app(struct opt *op, char *const *argv, char *const *envp)
{
	op->flags |= XTR_STARTCHLD;
	op->exename = argv[0];
	op->pid = fork();

	/* if we're a child, set correct PID  */
	if (op->pid == (pid_t) 0)
		op->pid = getpid();
	else
		return;

	/* child stops */
	(void) kill(op->pid, SIGSTOP);

	(void) execve(argv[0], argv, envp);

	/* if we're here, execve() failed */
}

/*
 * Check the process, is it dead or not. Assume procfs is working.
 */
int
xtr_isalive(pid_t pid)
{
	char fname[128];
	int ret = 0;		/* does not exits */

	(void) sprintf(fname, "/proc/%d", pid);

	if (access(fname, F_OK) == 0)
		++ret;

	return (ret);
}
