/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   API definition for xtrace utility, see README for details.
 *   this file is a part of xtrace project
 *
 * List of platform-specific functions, that have to be defined for every
 * platform specific module:
 *
 *   o xtr_init
 *   o xtr_start_app
 *   o xtr_stop_app
 *   o xtr_resume_app
 *   o xtr_get_call_data
 *   o xtr_read_mem
 *   o xtr_write_mem
 *
 * $Id: xtrace.h,v 1.21 2002/09/25 10:49:42 drozd Exp $
 */

#ifndef __XTRACE__H
#define __XTRACE__H

#include <sys/types.h>

#include <stdlib.h>
#include <unistd.h>
#include <stddef.h>

#include "config.h"
#include "llist.h"

typedef unsigned vaddr_t;

#define UNUSED(arg)	(arg) = (arg)

/*
 * Padding used to write syscalls return values
 */
#define PADDING			50

/* options */
struct opt {
	pid_t pid;		/* process's PID */
	unsigned flags;		/* see below for valid values */
	int scno;		/* syscall number */
	int signo;		/* signal number */
	vaddr_t stopva;		/* stop virtual address */
	int ncols;		/* characters printed */
	int memdumpsiz;		/* size of dumped mem content */
	int lock;		/* signal delivery lock */
	
	char *exename;		/* executable's full path name */
	char *flist;		/* comma-separated function list */
	char *slist;		/* comma-separated syscalls list */

	/* platform - specific stuff */
	void *data;			/* generic data */
	void (*fini)(struct opt *);	/* cleanup */
	void (*child)(struct opt *);	/* SIGCLD handler */
};


#define XTR_INSYSENTER	(1 << 0)	/* stopped in syscall entry */
#define XTR_INSYSEXIT	(1 << 1)	/* stopped in syscall exit */
#define XTR_INFENTER	(1 << 2)	/* stopped in fuction call enter */
#define XTR_INFEXIT	(1 << 3)	/* stopped in fuction exit */
#define XTR_SIGNAL	(1 << 4)	/* stopped because of a signal */
#define XTR_OTHER	(1 << 5)	/* stopped by some other reason */
#define XTR_SKIPPING	(1 << 6)	/* stopped due to singlestep */

#define XTR_BACKTRACE	(1 << 15)	/* dump per-thread backtrace */
#define XTR_TIMESTAMP	(1 << 16)	/* dump timestamp */
#define XTR_FUNCLIST	(1 << 17)	/* dump function list and exit */
#define XTR_DUMPMEM	(1 << 18)	/* dump memory content for parameters */
#define XTR_TRACESYS	(1 << 19)	/* trace system calls */
#define XTR_TRACEFUNC	(1 << 20)	/* trace functions */
#define XTR_DUMPTHREADS	(1 << 21)	/* dump per-thread info */
#define XTR_CHILDEXITED	(1 << 22)	/* child has called exit() */
#define XTR_STARTCHLD	(1 << 23)	/* xtrace has called start_app */
#define XTR_INTERRUPTED	(1 << 24)	/* syscall was interrupted by signal */
#define XTR_STEPPING	(1 << 25)	/* child single steps */
#define XTR_ATTACHED	(1 << 26)	/* attached to a child */
#define XTR_BACKTRACING	(1 << 27)	/* xtrace is doing backtrace */
#define XTR_INEXEC	(1 << 28)	/* execve() syscall performed by child */

#define BIT_SET(mask, bit) 	(mask |= bit)
#define BIT_UNSET(mask, bit) 	(mask &= ~bit)
#define BIT_ISSET(mask, bit) 	(mask & bit)


#define CLEAR_REASONS(op)					\
do {								\
	BIT_UNSET(op->flags, XTR_INFEXIT);			\
	BIT_UNSET(op->flags, XTR_INFENTER);			\
	BIT_UNSET(op->flags, XTR_INSYSEXIT);			\
	BIT_UNSET(op->flags, XTR_INSYSENTER);			\
	BIT_UNSET(op->flags, XTR_SIGNAL);			\
	BIT_UNSET(op->flags, XTR_OTHER);			\
	BIT_UNSET(op->flags, XTR_SKIPPING);			\
} while (0)

#define SET_REASON(op,r)					\
do { CLEAR_REASONS(op); BIT_SET(op->flags, r); } while (0)

#define SYMTAB_NEEDED(op) BIT_ISSET(op->flags, XTR_TRACEFUNC) ||	\
		BIT_ISSET(op->flags, XTR_BACKTRACING) ||	\
		BIT_ISSET(op->flags, XTR_BACKTRACE)


/*
 * lock field. This is necessary on Linux. wait() is called in xtr_stop_app,
 * and it can return *before* the signal handler was called. To synchronise
 * the signal handler and xtr_stop_app, simple lock is implemented. The value
 * of 0 means lock is locked, 1 means it is unlocked. The lock should be
 * locked by default and in xtr_stop_app, and unlocked in signal handler.
 */
#define UNLOCK(op)	do { (op)->lock = 1; } while (0)
#define LOCK(op)	do { while ((op)->lock == 0) ; (op)->lock = 0; } while (0)

/* symbol table */
struct symbol {
	vaddr_t va;		/* virtual address in debuggee address space */
	char *name;		/* symbol's name */
	unsigned flags;		/* see below for valid values */
};

#define SYM_REFERENCED		(1 << 0)	/* referenced from executable */
#define SYM_SETWATCH		(1 << 1)	/* watchpoint can be set */

/* In or Out parameter */
enum ArgFlow { IN, OUT };

/* parameter's type */
enum ArgType { HEX = 1, INT, STRING, SOCKADDR, MSGHDR, SIGALTSTACK,
	IOCTL, ARGV, ITIMER, FCNTL, TIMEVAL, TIMEZONE, IOVEC, STATFS,
	RTPRIO, STAT, RLIMIT, TIMESPEC, MODSTAT, KLDSTAT, SHEDPARAM,
	SIGSET, SIGACTION, KEVENT, FDSET };

struct Arg {
	enum ArgType type;
	enum ArgFlow inout;
};

#define MAXARGS		10
struct syscall {
	int no;					/* syscall's number */
	const char *name;			/* syscall's name */
	int nargs;				/* number of parameters */
	struct Arg args[MAXARGS];		/* MAXARGS parameters max */
};

/* function call data */
struct calldata {
	vaddr_t ip;		/* instruction pointer */
	vaddr_t retval;		/* return value */
	vaddr_t fp;		/* frame pointer */
	vaddr_t rta;		/* return address */
	vaddr_t ofp;		/* previous frame pointer */
	vaddr_t **argv;		/* array of arguments */
};

struct sharedlib {
	char *path;		/* full path to dependant library */
	vaddr_t base;		/* load address */
};

/* breakpoint structure */
struct bpt {
	struct llhead list;
	struct symbol *symp;
	vaddr_t va;
	unsigned long oldcode;	/* saved instruction */
	int codelen;
};

/* initialize tracer */
extern int xtr_init(struct opt *);

/* find symbol for given virtual address */
extern struct symbol *xtr_find_symbol(vaddr_t, ptrdiff_t *);

/* get list of dependant ELF binaries */
extern struct sharedlib **xtr_get_deps(pid_t);

/* build symbol table */
extern int xtr_symtab_build(struct opt *);

/* destroy symbol table */
extern void xtr_symtab_destroy(void);

/* get syscall's character string by number */
extern const struct syscall *xtr_lookup_syscall(int);

/* stop on event of interest */
extern int xtr_stop_app(struct opt *);

/* resume process */
extern int xtr_resume_app(struct opt *);

/* iterate over symbolint xtr_stop_app(const struct opt *);s */
extern struct symbol *xtr_get_next_sym(void);

/* read memory content from process's address space */
extern int xtr_read_mem(const struct opt *, vaddr_t, void *, int);

/* write memory content into process's address space */
extern int xtr_write_mem(const struct opt *, vaddr_t, const void *, int);

/* get call data by stack pointer */
extern int xtr_get_call_data(struct opt *, struct calldata *);

/* dump call */
extern void xtr_dump_call(struct opt *);

/* set breakpoint on function */
extern int xtr_set_bpt(struct opt *, vaddr_t);

/* clear breakpoint */
extern int xtr_clear_bpt(struct opt *, vaddr_t);

/* set breakpoints on functions */
extern void xtr_set_func_breakpoints(struct opt *op);

/* clear all breakpoints */
extern void xtr_clrbpts(struct opt *);

/* verify, does the process with specified PID exist or not */
extern int xtr_isalive(pid_t);

#ifdef __DEBUG__
#define dprintf(x) (void) fprintf x
#else
#define dprintf(x)
#endif /* __DEBUG__ */

#ifdef LINUX
#define waitlock(x) LOCK(x)
#else
#define waitlock(x)
#endif /* LINUX */

#ifdef I386
static const unsigned char BPT[] = { '\xcc' };
#endif /* I386 */

#endif /* __XTRACE__H */
